from meteor_reasoner.classes.interval import *
import copy


def contain_variable(entity):
    for term in entity:
        if term.type == "variable":
            return True
    return False


def contain_variable_after_replace(entity, context):
    for term in entity:
        if term.type == "variable" and term.name not in context:
            return True
    return False


def boundary_check(p, max_p):
    for i, _ in enumerate(p):
        if p[i] < max_p[i]:
            continue
        else:
            return False

    return True


def interval_merge(T):
    """
    A novel merging algorithm is implemented.
    Args:
        T (A list of list of Interval instances):

    Returns:
        A list of Interval instances.

    """

    T = list(copy.deepcopy(T))
    for i in range(len(T)):
        intervals = T[i]
        intervals = sorted(intervals, key=lambda t: t.left_value)
        T[i] = intervals

    T.sort(key=lambda item: item[0].left_value, reverse=True) # sorted by reverse according to the first interval's value

    max_pointers = [len(item) for item in T]
    pointers = [0] * len(T)
    result = set()
    while boundary_check(pointers, max_pointers):
        flag = True
        mover = T[0][pointers[0]]
        for i, _ in enumerate(pointers):
            while pointers[i] < max_pointers[i]:
                tmp = Interval.intersection(mover, T[i][pointers[i]])
                if tmp is None:
                    if T[i][pointers[i]].left_value == mover.right_value and mover.right_open and T[i][pointers[i]].left_open:
                        mover = tmp
                        for j in range(0, i):
                            while True:
                                if Interval.intersection(T[j][pointers[j]], T[i][pointers[i]]) is None and T[j][pointers[j]].left_value < T[i][pointers[i]].right_value:
                                    pointers[j] = pointers[j] + 1
                                    if pointers[j] == max_pointers[j]:
                                        break
                                else:
                                    break

                        flag = False
                        break

                    elif T[i][pointers[i]].left_value == mover.right_value and mover.right_open != T[i][pointers[i]].left_open:
                        mover = tmp
                        for j in range(0, i):
                            while True:
                                if Interval.intersection(T[j][pointers[j]], T[i][pointers[i]]) is None and T[j][pointers[j]].left_value < T[i][pointers[i]].right_value:
                                    pointers[j] = pointers[j] + 1
                                    if pointers[j] == max_pointers[j]:
                                        break
                                else:
                                    break

                        flag = False
                        break

                    elif T[i][pointers[i]].left_value > mover.right_value:
                        mover = tmp
                        for j in range(0, i):
                            while True:
                                if Interval.intersection(T[j][pointers[j]], T[i][pointers[i]]) is None and T[j][pointers[j]].left_value < T[i][pointers[i]].right_value:
                                    pointers[j] = pointers[j] + 1
                                    if pointers[j] == max_pointers[j]:
                                        break
                                else:
                                    break
                        flag = False
                        break

                    else:
                        if pointers[i] != max_pointers[i] - 1:
                            pointers[i] = pointers[i] + 1
                        else:
                            pointers[i] = pointers[i] + 1
                            flag = False
                            mover = None
                            break
                else:
                    mover = tmp
                    break

            if flag is False:
                break

        if mover is None:
            continue

        for i, _ in enumerate(pointers):
                if mover.right_value < T[i][pointers[i]].right_value:
                    T[i][pointers[i]].left_value = mover.right_value
                    T[i][pointers[i]].left_open = False
                elif mover.right_value == T[i][pointers[i]].right_value and mover.right_open and not T[i][pointers[i]].right_open:
                    T[i][pointers[i]].left_value = mover.right_value
                    T[i][pointers[i]].left_open = False
                else:
                    pointers[i] = pointers[i] + 1
        result.add(mover)
    return list(result)




from meteor_reasoner.classes.interval import *


def coalescing(old_intervals):
    if len(old_intervals) == 0:
        return old_intervals
    new_intervals = []
    old_intervals = sorted(old_intervals, key=lambda t: (t.left_value, t.left_open))
    i = 1
    mover = old_intervals[0]
    while i <= len(old_intervals)-1:
        tmp_interval = Interval.union(mover, old_intervals[i])
        if tmp_interval is None:
            # no intersection
            new_intervals.append(mover)
            mover = old_intervals[i]
        else:
            mover = tmp_interval
        i += 1
    new_intervals.append(mover)
    return new_intervals


def coalescing_d(D):
    """
    Merge two overlapped intervals into one interval.
    Args:
        D (a dictionary object): store facts.
    Returns:
    """
    for predicate in D:
        if type(D[predicate]) == list:
            old_intervals = D[predicate]
            if len(old_intervals) == 0:
                continue
            new_intervals = coalescing(old_intervals)
            D[predicate] = new_intervals

        else:
            for entity, old_intervals in D[predicate].items():
                old_intervals = D[predicate][entity]
                if len(old_intervals) == 0:
                    continue
                new_intervals = coalescing(old_intervals)
                D[predicate][entity] = new_intervals


from itertools import chain, combinations
from meteor_reasoner.automata.window import *
from meteor_reasoner.automata.satisfy import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.automata.guess import *
from meteor_reasoner.materialization.apply import *


class BuchiAutomata:
    def __init__(self, D, program, unbounded_literals, p_literals, involved_predicates, automata_predicates,left_dict,
                 right_dict, x, z, gcd, points):
        self.D = D
        self.program = program
        self.x = x
        self.z = z
        self.gcd = gcd
        self.points = points
        self.p_literals = p_literals

        self.unbounded_literals = unbounded_literals
        self.window_ruler_interval_list = []

        self.involved_predicates = involved_predicates
        self.automata_predicates = automata_predicates

        self.left_window_pattern_dict = left_dict
        self.right_window_pattern_dict = right_dict

        self.window_must_include_static = defaultdict(list)
        self.variable_literals = set()

        self.left_conditions = self.get_conditions("left")
        self.right_conditions = self.get_conditions("right")

    def get_conditions(self, d):
        F = set()
        for literal in self.unbounded_literals:
            if d == "left":
                if isinstance(literal, BinaryLiteral):
                    if literal.operator.interval.right_value == float("inf") and literal.operator.name == "Since":
                        F.add(literal)
                else:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and (
                            literal.operators[0].interval.right_value == float("inf")) and literal.operators[0].name == "Boxminus":
                        F.add(literal)

            else:
                if isinstance(literal, BinaryLiteral):
                    if literal.operator.interval.right_value == float("inf") and literal.operator.name == "Until":
                        F.add(literal)
                else:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and (
                            literal.operators[0].interval.right_value == float("inf")) and literal.operators[0].name == "Boxplus":
                        F.add(literal)
        return F

    def build_prior(self, must_literals):
        initial_window_len, initial_window_ruler_intervals = get_initial_ruler_intervals(self.points,
                                                                                         left_border=-2 * self.x - 2 * self.z,
                                                                                         right_border=2 * self.x + 2 * self.z,
                                                                                         gcd=self.gcd)
        self.window_ruler_interval_list = initial_window_ruler_intervals
        for ruler_interval in initial_window_ruler_intervals:
            self.window_must_include_static[ruler_interval] = []

        for ruler_interval in initial_window_ruler_intervals:
            for literal in must_literals:
                if interval_intesection_intervallist(ruler_interval, must_literals[literal]):
                    self.window_must_include_static[ruler_interval].append(literal)
                if isinstance(literal, BinaryLiteral):
                    if literal.left_literal.get_predicate() in self.automata_predicates or literal.right_literal.get_predicate() in self.automata_predicates:
                        self.variable_literals.add(literal)
                else:
                    if literal.get_predicate() in self.automata_predicates:
                        self.variable_literals.add(literal)

            for predicate in self.involved_predicates:
                if type(self.D[predicate]) == list:
                    if interval_intesection_intervallist(ruler_interval, self.D[predicate]):
                        self.window_must_include_static[ruler_interval].append(Atom(predicate))
                    if predicate in self.automata_predicates:
                        self.variable_literals.add(Atom(predicate))
                else:
                    for entity in self.D[predicate]:
                        if interval_intesection_intervallist(ruler_interval, self.D[predicate][entity]):
                            self.window_must_include_static[ruler_interval].append(
                                Atom(predicate, entity))
                        if predicate in self.automata_predicates:
                            self.variable_literals.add(Atom(predicate, entity))

    def get_unbounded_literals_for_window(self, atom, ruler_interval, direction):
        """
        Return a list of possible unbounded literals for the given ruler_interval based
        on the pre-materialized results in [-2x-2z, 2x+2z]
        Args:
            atom (Atom instance):
            ruler_interval (Interval instance):
            direction (str): ``left'' or ``right''

        Returns:
            list of Literal instances

        """
        if ruler_interval not in self.window_must_include_static:
            return []
        unbounded_literals = []
        start_i = self.window_ruler_interval_list.index(ruler_interval)
        if direction == "left":
            for i in range(start_i, -1, -1):
                next_ruler_interval = self.window_ruler_interval_list[i]
                if atom not in self.window_must_include_static[next_ruler_interval]:
                    break
            else:
                open_operator = Operator("Boxminus", Interval(0, float("inf"), True, True))
                closed_operator = Operator("Boxminus", Interval(0, float("inf"), False, True))
                unbounded_literals.append(Literal(atom, [open_operator]))
                unbounded_literals.append(Literal(atom, [closed_operator]))
        else:
            for i in range(start_i, len(self.window_ruler_interval_list)):
                next_ruler_interval = self.window_ruler_interval_list[i]
                if atom not in self.window_must_include_static[next_ruler_interval]:
                    break
            else:
                open_operator = Operator("Boxplus", Interval(0, float("inf"), True, True))
                closed_operator = Operator("Boxplus", Interval(0, float("inf"), False, True))
                unbounded_literals.append(Literal(atom, [open_operator]))
                unbounded_literals.append(Literal(atom, [closed_operator]))

        return unbounded_literals

    def get_literals_for_ruler_interval(self, ruler_interval, direction, last_ruler_interval_literals=None):
        """
        Return a list of Atom or Literal instances for a given ruler interval
        Args:
            ruler_interval (an Interval instance):

        Returns:
            list of Atoms or Literal instances

        """
        static_must_literals = set(self.window_must_include_static[ruler_interval])
        unbounded_must_literals = set()
        if last_ruler_interval_literals is not None:
            if direction == "left":
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 0 and literal.operators[ 0].name == "Boxminus" and\
                            literal.operators[0].interval.right_value == float("inf"):
                        unbounded_must_literals.add(literal)
                        unbounded_must_literals.add(literal.atom)
            else:
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 0 and literal.operators[ 0].name == "Boxplus" and \
                            literal.operators[0].interval.right_value == float("inf"):

                        unbounded_must_literals.add(literal)
                        unbounded_must_literals.add(literal.atom)


        for com_literals in chain.from_iterable(combinations(self.variable_literals, r)
                                                   for r in range(len(self.variable_literals) + 1)):

            yield set(com_literals) | static_must_literals | unbounded_must_literals

    def get_literals_for_ruler_interval_with_visited(self, ruler_interval, direction, visited, last_ruler_interval_literals=None):
        """
        Return a list of Atom or Literal instances for a given ruler interval
        Args:
            ruler_interval (an Interval instance):

        Returns:
            list of Atoms or Literal instances

        """
        static_must_literals = self.window_must_include_static[ruler_interval]
        unbounded_must_literals = []
        for literal in static_must_literals:
            if isinstance(literal, Atom) and literal not in visited:
                if direction == "left":
                    res = self.get_unbounded_literals_for_window(literal, ruler_interval, "left")
                    if len(res) > 0:
                        visited[literal] = 1
                        unbounded_must_literals += res
                else:
                    res = self.get_unbounded_literals_for_window(literal, ruler_interval, "right")
                    if len(res) > 0:
                        visited[literal] = 1
                        unbounded_must_literals += res

        if last_ruler_interval_literals is not None:
            if direction == "left":
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and literal.operators[0].name == "Boxminus" and \
                            literal.operators[0].interval.right_value == float("inf"):
                        unbounded_must_literals.append(literal)
                        unbounded_must_literals.append(literal.atom)
            else:
                for literal in last_ruler_interval_literals:
                    if isinstance(literal, Literal) and len(literal.operators) == 1 and literal.operators[0].name == "Boxplus" and \
                            literal.operators[0].interval.right_value == float("inf"):
                        unbounded_must_literals.append(literal)
                        unbounded_must_literals.append(literal.atom)

        for com_literals in chain.from_iterable(combinations(self.variable_literals, r)
                                                   for r in range(len(self.variable_literals) + 1)):

            yield list(com_literals) + static_must_literals + unbounded_must_literals

    def get_literals_for_window(self, window_ruler_interval_list,  visited, direction):
        def helper(index, result):
            if index == len(window_ruler_interval_list):
                window_ruler_intervals_literals = dict()
                for key, value in result.items():
                    window_ruler_intervals_literals[key] = value

                yield AutomataWindow(window_ruler_interval_list, window_ruler_intervals_literals,
                                     self.left_window_pattern_dict, self.right_window_pattern_dict)
                return

            else:
                last_interval_literals = None
                if index != 0:
                    last_interval_literals = result[window_ruler_interval_list[index-1]]

                for com_literals in self.get_literals_for_ruler_interval_with_visited(window_ruler_interval_list[index], direction, visited, last_interval_literals):
                    yield from helper(index+1, {**result, **{window_ruler_interval_list[index]: list(com_literals)}})

        yield from helper(0, {})

    def check_inconsistency(self):
        """
        This function check whether the given D and Program are consistency by running Buichi Automata.
        Returns:
           boolean
        """
        initial_window_len, initial_window_ruler_intervals =get_initial_ruler_intervals(
            [- self.x - self.z, -self.x + self.z])
        for l_min, r_min in self.search_window(initial_window_ruler_intervals):
            if self.search(l_min, "left", self.F_left):
                if self.search(r_min, "right", self.F_right):
                    return True
        return False

    def search_window(self, initial_window_ruler_intervals):
        """
        Instead of directly checking the window=[-x-z, -x+z, we start from
        the minimum window=[-x-z, -x+z], and then move in a rightward direction
        at a gcd step until the window reach to x+z. Then return two windows:
        lmin=[-x-z, -x+z], rmin=[x-z, x+z]
        Args:
            initial_window_ruler_intervals (list of Interval instances):

        Returns:
            (AutomataWindow instance, AutomataWindow instance)

        """
        visited = dict()
        def guess_help(w, w0):
            while w.ruler_intervals[-1] != Interval(self.x+self.z, self.x+self.z, False, False):
                last_interval_literals = w.ruler_intervals_literals[w.ruler_intervals[-1]]
                w.move_right()
                last_interval = w.ruler_intervals[-1]
                for com_literals in self.get_literals_for_ruler_interval_with_visited(last_interval, "right", visited, last_interval_literals):
                    w.ruler_intervals_literals[last_interval] = com_literals
                    if not self.check_satisfy(w):
                        return None
                    else:
                        break
            return (w0, w)

        # This is a recursive version
        def helper(stack):

            if len(stack) > 0 and stack[-1].ruler_intervals[-1] == Interval(self.x+self.z, self.x+self.z, False, False):
                yield stack[0], stack[-1]
                return
            last_w = stack[-1]
            new_w = copy.deepcopy(last_w)
            last_interval_literals = new_w.ruler_intervals_literals[new_w.ruler_intervals[-1]]
            new_w.move_right()
            last_interval = new_w.ruler_intervals[-1]

            for com_literals in self.get_literals_for_ruler_interval_with_visited(last_interval, "right", visited, last_interval_literals):
                new_w.ruler_intervals_literals[last_interval] += com_literals
                if self.check_satisfy(new_w):
                    yield from helper(stack+[new_w])
        flag = True
        for w in self.get_literals_for_window(initial_window_ruler_intervals, visited, "right"):
            if self.check_satisfy(w):
                if flag:
                    res = guess_help(copy.deepcopy(w), copy.deepcopy(w))
                    flag = False
                    if res is not None:
                        yield res[0], res[1]
                    else:
                        yield from helper([w])

                else:
                    yield from helper(w, copy.deepcopy(w))

    def consistency_check(self):
        """
        Returns:
           boolean
        """
        initial_window_len, initial_window_ruler_intervals = get_initial_ruler_intervals(self.points,
            left_border=- self.x - self.z, right_border= -self.x + self.z, gcd=self.gcd)

        for l_min, r_min in self.search_window(initial_window_ruler_intervals):
            left_flag = self.search(l_min, "left", self.left_conditions)
            if left_flag:
                right_flag = self.search(r_min, "right", self.right_conditions)
                if right_flag:
                    return True
        return False

    def search(self, W_0, d, F):
        T = list()
        R = set()
        label = defaultdict(set)
        T.append(W_0)

        while len(T) > 0:
            W = T[-1]
            increase_flag = True
            while increase_flag:
                for tmp_interval in W.ruler_intervals:
                    if len(W.ruler_intervals_literals[tmp_interval]) != 0:
                        break
                else:
                    return True
                if d == "left":
                    W_prime = copy.deepcopy(W)
                    last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                    W_prime.move_left()
                    first_interval = W_prime.ruler_intervals[0]
                    for com_literals in self.get_literals_for_ruler_interval(first_interval,d, last_interval_literals):
                        W_prime.ruler_intervals_literals[first_interval] = com_literals
                        if self.check_satisfy(W_prime) and W_prime not in T + list(R):
                            T.append(W_prime)
                            label[W_prime] = set()
                            W = W_prime
                            break
                    else:
                        increase_flag = False
                else:
                    W_prime = copy.deepcopy(W)
                    last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                    W_prime.move_right()
                    last_interval = W_prime.ruler_intervals[-1]
                    for com_literals in self.get_literals_for_ruler_interval(last_interval, d, last_interval_literals):
                        W_prime.ruler_intervals_literals[last_interval] = com_literals
                        if self.check_satisfy(W_prime) and W_prime not in T + list(R):
                            T.append(W_prime)
                            label[W_prime] = set()
                            W = W_prime
                            break
                        else:
                            increase_flag = False
                            break
                    else:
                        increase_flag = False
            subset_conditions = set()
            if (W in label and len(label[W]) != 0) or window_contain_accepting_conditions(F, W, subset_conditions, d):
                label[W] = label[W].union(subset_conditions)
                if label[W] == F:
                    return True
                self.propagate([W], label[W], T, R, label, d)
                if label[W] == F:
                    return True
            R.add(W)
            T.pop()

        return False

    def propagate(self, N, P, T, R, label, d):
        while len(N) != 0:
            W = N.pop()
            if d == "left":
                W_prime = copy.deepcopy(W)
                last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                W_prime.move_left()

                first_interval = W_prime.ruler_intervals[0]
                for subset_literal in self.get_literals_for_ruler_interval(first_interval, d, last_interval_literals):
                    W_prime.ruler_intervals_literals[first_interval] = subset_literal

                    if self.check_satisfy(W_prime) and W_prime in T + list(R):
                        if self.can_contribute(P, label[W_prime]):
                            N.append(W_prime)
                            label[W_prime] = label[W_prime].union(P)
            else:
                W_prime = copy.deepcopy(W)
                last_interval_literals = W_prime.ruler_intervals_literals[W_prime.ruler_intervals[-1]]
                W_prime.move_right()

                last_interval = W_prime.ruler_intervals[-1]
                for subset_literal in self.get_literals_for_ruler_interval(last_interval, d, last_interval_literals):
                    W_prime.ruler_intervals_literals[last_interval] = subset_literal
                    if self.check_satisfy(W_prime) and W_prime in T + list(R):
                        if self.can_contribute(P, label[W_prime]):
                            N.append(W_prime)
                            label[W_prime] = label[W_prime].union(P)

    def can_contribute(self, a, b):
        for item in a:
            if item not in b:
                return True
        return False

    def check_satisfy(self, w):
        if not check_satisfy_dataset(w, self.D):
            return False
        ruler_intervals_literals = w.ruler_intervals_literals
        ruler_intervals = w.ruler_intervals

        guess_D = defaultdict(lambda: defaultdict(list))
        binary_literals = []

        for ruler_interval in ruler_intervals:
            for literal in ruler_intervals_literals[ruler_interval]:
                if isinstance(literal, Atom):
                    if literal.get_entity() is None:
                        if literal.get_predicate() not in guess_D:
                            guess_D[literal.get_predicate()] = [ruler_interval]
                        else:
                            guess_D[literal.get_predicate()].append(ruler_interval)
                    else:
                        guess_D[literal.get_predicate()][literal.get_entity()].append(ruler_interval)
                elif isinstance(literal, BinaryLiteral):
                    binary_literals.append(literal)

                elif isinstance(literal, Literal):
                    if literal.operators[0].name == "Boxplus":
                        interval = Interval.add(ruler_interval, literal.operators[0].interval)
                        if literal.get_entity() is None:
                            if literal.get_predicate() not in guess_D:
                                guess_D[literal.get_predicate()] = [interval]
                            else:
                                guess_D[literal.get_predicate()].append(interval)
                        else:
                            if literal.get_predicate() not in guess_D:
                                guess_D[literal.get_predicate()][literal.get_entity()] = [interval]
                            else:
                                guess_D[literal.get_predicate()][literal.get_entity()].append(interval)

                    elif literal.operators[0].name == "Boxminus":
                        interval = Interval.sub(ruler_interval, literal.operators[0].interval)
                        if literal.get_entity() is None:
                            if literal.get_predicate() not in guess_D:
                                guess_D[literal.get_predicate()] = [interval]
                            else:
                                guess_D[literal.get_predicate()].append(interval)
                        else:
                            if literal.get_predicate() not in guess_D:
                                guess_D[literal.get_predicate()][literal.get_entity()] = [interval]
                            else:
                                guess_D[literal.get_predicate()][literal.get_entity()].append(interval)

                else:
                    raise Exception("Please convert Diamondminus or Diamondplus to Boxminus or Boxplus")


        if len(binary_literals) != 0:
            guess_points, left_right_literals = generate_guess_points(w, binary_literals, ruler_intervals, self.z, self.gcd)
            # guess existing intervals
            for literal in left_right_literals:
                if literal.get_predicate() in self.D and literal.get_entity() is None:
                    guess_D[literal.get_predicate()] = D[literal.get_predicate()]
                elif literal.get_predicate() in self.D and literal.get_entity() in self.D[literal.get_predicate()]:
                    guess_D[literal.get_predicate()][literal.get_entity()] = self.D[literal.get_predicate()][literal.get_entity()]
                else:
                    # guess an interval from guess_points
                    pass

        guess_D = coalescing_d(guess_D)
        for literal in self.p_literals:
            T = apply(literal, guess_D)
            if len(T) == 0:
                for ruler_interval in ruler_intervals:
                    if literal in ruler_intervals_literals[ruler_interval]:
                        return False
            else:
                for ruler_interval in ruler_intervals:
                    if interval_intesection_intervallist(ruler_interval, T) and literal not in \
                            ruler_intervals_literals[ruler_interval]:
                        return False

        prgram_flag = check_satisfy_program(w, program=self.program)
        return prgram_flag
































from meteor_reasoner.automata.utils import *
from meteor_reasoner.classes.atom import Atom
import copy


def check_satisfy_dataset(w, D, involved_predicates=[]):
    """
    This function is to check whether all facts in ``D'' have been installed in each of ruler intervals
    of the given Window ``w'' if facts in ruler intervals holds in ``D''.
    Args:
        w (a Window instance):
        D (dictionary of dictionary object): contain all facts
        involved_predicates (a list of str): contain all predicates that are needed to be checked.

    Returns:
        boolean

    """
    for ruler_interval in w.ruler_intervals:
        for predicate in involved_predicates:
            if type(D[predicate]) == list:
                interval_list = D[predicate]
                if interval_intesection_intervallist(ruler_interval, interval_list) and Atom(predicate) not \
                        in w.get_ruler_intervals_literals()[ruler_interval]:
                    return False
            else:
                for entity, interval_list in D[predicate].items():
                    if interval_intesection_intervallist(ruler_interval, interval_list) and Atom(predicate) not \
                            in w.get_ruler_intervals_literals()[ruler_interval]:
                        return False
    return True


def check_satisfy_program(w, program):
    """
    For each rule in the program, this function checks whether the head exists if all literals(or atoms) in the body exists
    in  each ruler interval of the given Window ``w'' .
    Args:
        w (a Window instance):
        program (a list of rules):

    Returns:
        boolean.

    """
    ruler_intervals_literals = w.ruler_intervals_literals
    for ruler_interval in w.ruler_intervals:
        for rule in program:
            if not ground_rule(rule, ruler_intervals_literals[ruler_interval]):
                return False
    return True


def get_matched_literals(literal, installed_literals):
    """
    Given a literal, this function is to return all literals in ``installed_literals'' that have same predicate and operator(s).
    Args:
        literal (an Atom, Literal or BiLiteral instance):
        installed_literals (a list of Atom, Literal or BiLiteral instances in a ruler interval):

    Returns:
        A list of matched Atom or Literal or BiLiteral instances.

    """
    predicate_matched_literals = []
    for item in installed_literals:
        if isinstance(item, BinaryLiteral) and isinstance(literal, BinaryLiteral):
            if literal.left_literal.get_predicate() == item.left_literal.get_predicate() and literal.right_literal.get_predicate() == item.right_literal.get_predicate():
                if literal.operator == item.operator:
                    predicate_matched_literals.append(item)
        elif isinstance(item, Literal) and isinstance(literal, Literal):
            if literal.get_predicate() == item.get_predicate() and literal.operators == item.operators:
                predicate_matched_literals.append(item)
        elif isinstance(item, Atom) and isinstance(literal, Atom):
            if literal.get_predicate() == item.get_predicate():
                predicate_matched_literals.append(item)

    return predicate_matched_literals


def ground_literal(literal, context, installed_literals):
    """
    Check whether the literal Ground the
    Args:
        literal:
        context:
        installed_literals:

    Returns:

    """
    if isinstance(literal, BinaryLiteral):
        for item in get_matched_literals(literal, installed_literals):
            tmp_context = dict()
            literal_left_entity = literal.left_literal.get_entity()
            item_left_entity = item.left_literal.get_entity()
            left_valid_flag = True

            if literal.left_literal.get_predicate() not in ["Top"]:
                if literal_left_entity is not None:
                    for i in range(len(literal_left_entity)):
                        if literal_left_entity[i].type == "constant":
                            if literal_left_entity[i] != item_left_entity[i]:
                                left_valid_flag = False
                                break
                        else:
                            if literal_left_entity[i].name not in context:
                                tmp_context[literal_left_entity[i].name] = item_left_entity[i]
                            else:
                                if literal_left_entity[i].name != context[literal_left_entity[i].name]:
                                    left_valid_flag = False
                                    break

            if left_valid_flag:
                literal_right_entity = literal.right_literal.get_entity()
                item_right_entity = item.right_literal.get_entity()
                right_valid_flag = True

                if literal.right_literal.get_predicate() not in ["Top"]:
                    if literal_right_entity is not None:
                        for i in range(len(literal_right_entity)):
                            if literal_right_entity[i].type == "constant":
                                if literal_right_entity[i] != item_right_entity[i]:
                                    right_valid_flag = False
                                    break
                            else:
                                if literal_right_entity[i].name not in context:
                                    tmp_context[literal_right_entity[i].name] = item_right_entity[i]
                                else:
                                    if literal_right_entity[i].name != context[literal_right_entity[i].name]:
                                        right_valid_flag = False
                                        break

                if right_valid_flag:
                    yield item, tmp_context
    else:
        entity = literal.get_entity()
        tmp_context = dict()
        valid_flag = True
        for item in get_matched_literals(literal, installed_literals):
            if entity is not None:
                for i in range(len(entity)):
                    if entity[i].type == "constant":
                        if entity[i] != item.get_entity()[i]:
                            valid_flag = False
                            break
                    else:
                        if entity[i].name not in context:
                            tmp_context[entity[i].name] = item.get_entity()[i]
                        else:
                            if item.get_entity()[i] != context[entity[i].name]:
                                valid_flag = False
                                break
            if valid_flag:
                yield item, tmp_context


def ground_rule(rule, installed_literals):
    """
    Grounding the rule and then check whether the grounded head exists when all grounded literals in the body exists.
    Args:
        rule (a Rule instance):
        installed_literals (a list of Literal or Atom instances):
    Returns:
        boolean
    """
    head_predicate = rule.head.get_predicate()
    head_entity = rule.head.get_entity()
    body_literals = rule.body

    def ground_body(global_literal_index, context, flag):
        if global_literal_index == len(body_literals):
            if head_predicate == "Bottom":
                flag[0] = False
                return
            if head_entity is None:
                if rule.head not in installed_literals:
                    flag[0] = False
                return

            ground_head_entity = []
            for i, term in enumerate(head_entity):
                if term.type == "constant":
                    ground_head_entity.append(term)
                else:
                    ground_head_entity.append(context[term.name])

            ground_head_entity = tuple(ground_head_entity)
            ground_head = copy.deepcopy(rule.head)
            ground_head.set_entity(ground_head_entity)
            if ground_head not in installed_literals:
                flag[0] = False
                return
        else:
            for _, tmp_context in ground_literal(body_literals[global_literal_index], context, installed_literals):
                if flag[0]:
                    ground_body(global_literal_index+1, {**context, **tmp_context}, flag)

    flag = [True]
    ground_body(0, {}, flag)
    return flag[0]








from meteor_reasoner.graphutil.graph_strengthening import *
from meteor_reasoner.graphutil.multigraph import *
from meteor_reasoner.classes import *
import copy


def test_finite_materialisability():
    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False)),
                                                                    Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])
    literal_c = BinaryLiteral(copy.deepcopy(literal_a), copy.deepcopy(literal_b),
                              Operator("Until", Interval(4, 5, False, False)))

    body = [literal_a, literal_b, literal_c]
    rule = Rule(head, body)
    rules = transformation([rule])
    print(rules[0])
    paris = construct_pair(rules)
    G = TemporalDependencyGraph()
    for pair in paris:
        G.add_edge(pair[0], pair[1], [pair[2], pair[3]])
    print(rule, ":")
    print(G.is_cyclic())


test_finite_materialisability()




from setuptools import setup, find_packages
from os import path
import sys

from io import open
here = path.abspath(path.dirname(__file__))
sys.path.insert(0, path.join(here, 'meteor_reasoner'))
# from version import __version__
#
# print('version')
# print(__version__)

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(name='meteor_reasoner',
      version="1.0.5",
      description='A Metric Temporal Reasoner',
      url='https://github.com/wdimmy/MeTeoR',
      author='Dingmin Wang',
      author_email='dingmin.wang@cs.ox.ac.uk',
      keywords=['DatalogMTL', 'Knowledge Representation', 'Temporal Reasoning'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      install_requires = [
        'outdated>=0.2.0'
      ],
      license='MIT',
      packages=find_packages(exclude=["data"]),
      extras_require={
          "interactive" : ['matplotlib>=2.2.0', 'jupyter']
      },
      setup_requires=['pytest-runner'],
      tests_require=["pytest"],
      include_package_data=True,
      classifiers=[
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
    ],
)



class Operator:
    """
    The operator class is designed for storing six MTL operators, e.g., Boxminus[1,2],
    Boxplus(1,2), Diamondplus[3,4], Diamondminus[1,2], Since[1,2], Until[2,3]
    """
    def __init__(self, name, interval):
        """
        Args:
            name (str): Boxminus, Boxplus, Diamondplus, Diamondminus, Since, Until
            interval (Interval class):
        """
        if name not in ["Boxminus", "Boxplus", "Diamondplus", "Diamondminus", "Since", "Until"]:
            raise ValueError("Only support one of six operators (Boxminus, Boxplus, Diamondplus, "
                             "Diamondminus, Since, Until)!")

        self.name = name
        self.interval = interval

    def __eq__(self, other):
        if isinstance(other, Operator) and self.name == other.name and self.interval == other.interval:
            return True

        return False

    def __str__(self):
        return self.name + str(self.interval)


class Literal:
    """
    Literal class is for storing a literal consisting of one atom with one or multiple MTL operators,
    e.g., Boxminus[1,2]Professor(X)
    """
    def __init__(self, atom, operators=[]):
        """

        Args:
            atom (Atom class):
            operators (A list of Operator class):
        """
        self.atom = atom
        self.operators = operators

    def get_predicate(self):
        return self.atom.predicate

    def get_entity(self):
        return self.atom.entity

    def get_op_name(self): # always return the name of the first operator
        if len(self.operators) != 0:
            return self.operators[0].name
        else:
            return None

    def set_entity(self, entity):
        self.atom.entity = entity

    def __eq__(self, other):
        if isinstance(other, Literal):
            if self.atom == other.atom and self.operators == other.operators:
                return True
        return False

    def __str__(self):
        return "".join([str(operator) for operator in self.operators]) + str(self.atom)

    def __hash__(self):
        hash_value = hash("Literal" + str(self.atom) + "".join([str(operator) for operator in self.operators]))
        return  hash_value


class BinaryLiteral:
    """
    The BinaryLiteral class is used to store binary literal, in which there are two literal separated by
    Until or Since Operator, e.g. A(X)Since[1,2]B(X,Y)
    """
    def __init__(self, left_literal, right_literal, operator):
        """

        Args:
            left_atom (Atom class):
            right_atom (Atom class):
            operator (Operator class):
        """
        self.left_literal = left_literal
        self.right_literal = right_literal
        self.operator = operator

    def get_op_name(self):
        return self.operator.name

    def set_entity(self, entities):
        if self.left_literal.get_predicate() == "Top":
            self.left_literal.set_entity(entities[0])
        elif self.right_literal.get_predicate() == "Top":
            self.right_literal.set_entity(entities[0])
        else:
            self.left_literal.set_entity(entities[0])
            self.right_literal.set_entity(entities[1])

    def __eq__(self, other):
        if isinstance(other, BinaryLiteral):
            return self.left_literal == other.left_literal and self.right_literal == other.right_literal and \
                   self.operator == other.operator
        return False

    def __str__(self):
        return str(self.left_literal) + str(self.operator) + str(self.right_literal)

    def __hash__(self):
        return hash("BiLiteral:" + str(self.left_literal) + str(self.operator) + str(self.right_literal))










from meteor_reasoner.classes.rule import *
from meteor_reasoner.materialization.ground import *
from meteor_reasoner.classes.term import Term
from collections import defaultdict
from meteor_reasoner.materialization.index_build import build_index
from meteor_reasoner.utils.operate_dataset import print_dataset
from meteor_reasoner.materialization.coalesce import coalescing_d
from decimal import Decimal


def inclusion_check(D, rule, i, context, records, single=False):
    head_literal = copy.deepcopy(rule.head)

    if isinstance(head_literal, BinaryLiteral):
        left_head_literal = head_literal.left_literal
        right_head_literal = head_literal.right_literal

        left_predicate = left_head_literal.get_predicate()
        right_predicate = right_head_literal.get_predicate()

        if left_predicate not in D or right_predicate not in D:
            return False

        left_head_entity = left_head_literal.get_entity()
        right_head_entity = right_head_literal.get_entity()
        left_head_replaced_entity = []
        for term in left_head_entity:
            if term.type == "variable":
                term.type = "constant"
                term.name = context[term.name]
                left_head_replaced_entity.append(term)
            else:
                left_head_replaced_entity.append(term)
        right_head_replaced_entity = []
        for term in right_head_entity:
            if term.type == "variable":
                term.type = "constant"
                term.name = context[term.name]
                right_head_replaced_entity.append(term)
            else:
                right_head_replaced_entity.append(term)
        left_head_replaced_entity = tuple(left_head_replaced_entity)
        right_head_replaced_entity = tuple(right_head_replaced_entity)
        head_literal.left_literal.set_entity(left_head_replaced_entity)
        head_literal.right_literal.set_entity(right_head_replaced_entity)
        if left_head_replaced_entity not in D[left_predicate] or right_head_replaced_entity not in D[right_predicate]:
            return False

    else:
        head_predicate = head_literal.get_predicate()
        if head_predicate not in D:
            return False
        head_entity = head_literal.get_entity()
        head_replaced_entity = []
        for term in head_entity:
            if term.type == "variable":
                term.type = "constant"
                term.name = context[term.name]
                head_replaced_entity.append(term)
            else:
                head_replaced_entity.append(term)
        head_replaced_entity = tuple(head_replaced_entity)
        head_literal.set_entity(tuple(head_replaced_entity))
        if head_replaced_entity not in D[head_predicate]:
            return False

    prefix_literals = [head_literal]
    if single:
        start = i
    else:
        start = 0

    for literal in rule.body[start:i+1]:
        ground_literal = copy.deepcopy(literal)
        if isinstance(ground_literal, BinaryLiteral):
            left_head_literal = ground_literal.left_literal
            right_head_literal = ground_literal.right_literal
            left_head_entity = left_head_literal.get_entity()
            right_head_entity = right_head_literal.get_entity()
            left_head_replaced_entity = []
            for term in left_head_entity:
                if term.type == "variable":
                    term.type = "constant"
                    term.name = context[term.name]
                    left_head_replaced_entity.append(term)
                else:
                    left_replaced_entity.append(term)
            right_replaced_entity = []
            for term in right_head_entity:
                if term.type == "variable":
                    term.type = "constant"
                    term.name = context[term.name]
                    right_replaced_entity.append(term)
                else:
                    right_head_replaced_entity.append(term)

            left_replaced_entity = tuple(left_replaced_entity)
            right_replaced_entity = tuple(right_replaced_entity)
            ground_literal.left_literal.set_entity(left_replaced_entity)
            ground_literal.right_literal.set_entity(right_replaced_entity)

        else:
            body_entity = ground_literal.get_entity()
            body_replaced_entity = []
            for term in body_entity:
                if term.type == "variable":
                    term.type = "constant"
                    term.name = context[term.name]
                    body_replaced_entity.append(term)
                else:
                    body_replaced_entity.append(term)
            body_replaced_entity = tuple(body_replaced_entity)
            ground_literal.set_entity(body_replaced_entity)

        if tuple(prefix_literals+[ground_literal]) in records:
            return True
        prefix_literals.append(ground_literal)

    head_intervals = reverse_apply(head_literal, D)
    flag = False
    body_intervals = [Interval(Decimal("-inf"), Decimal("inf"), True, True)]
    for i, literal in enumerate(prefix_literals):
        if i == 0:
            continue # at least two literals
        t = apply(literal, D)
        body_intervals = Interval.list_union(body_intervals, t)
        if len(body_intervals) == 0 or Interval.list_inclusion(body_intervals, head_intervals):
            flag = True
            records[tuple(prefix_literals[:i+1])] = True # can add some permutation, TOD
            break
    return flag



def optimized_seminaive_join(rule, D,  delta_old, delta_new,  D_index=None, records= {}, non_predicates=dict()):
    """
    This function implement the join operator when variables exist in the body of the rule.
    Args:
        rule (a Rule instance):
        Delta (a dictionary object): store new materialized results.
        D (a dictionary object): a global variable storing all facts.

    Returns:
        None
    """
    head_entity = rule.head.get_entity()
    head_predicate = rule.head.get_predicate()
    literals = rule.body + rule.negative_body

    def ground_body(global_literal_index, visited, delta, context):
        if global_literal_index == len(literals):
            T = []
            for i in range(len(rule.body)):
                grounded_literal = copy.deepcopy(literals[i])
                if isinstance(grounded_literal, BinaryLiteral):
                    grounded_literal.set_entity(delta[i])
                else:
                    if grounded_literal.get_predicate() not in ["Bottom", "Top"]:
                        grounded_literal.set_entity(delta[i][0])
                if i == visited:
                   t = apply(grounded_literal, delta_old)
                else:
                   t = apply(grounded_literal, D)
                if len(t) == 0:
                    break
                else:
                    T.append(t)
            n_T = []
            for i in range(len(rule.body), len(literals)):
                grounded_literal = copy.deepcopy(literals[i])
                if isinstance(grounded_literal, BinaryLiteral):
                    grounded_literal.set_entity(delta[i])
                else:
                    if grounded_literal.get_predicate() not in ["Bottom", "Top"]:
                        grounded_literal.set_entity(delta[i][0])
                t = apply(grounded_literal, D)
                if len(t) == 0:
                    break
                else:
                    n_T.append(t)

            if len(n_T) > 0 and len(n_T) == len(rule.negative_body_atoms):
                n_T = interval_merge(T)
            else:
                n_T = []

            if head_entity is not None:
                replaced_head_entity = []
                for term in head_entity:
                    if term.type == "constant":
                        replaced_head_entity.append(term)
                    else:
                        if term.name not in context:
                            raise ValueError("Head variables in Rule({}) do not appear in the body".format(str(rule)))
                        else:
                            new_term = Term.new_term(context[term.name])
                            replaced_head_entity.append(new_term)
                replaced_head_entity = tuple(replaced_head_entity)
            else:
                replaced_head_entity = head_entity

            if len(T) == len(literals):
                T = interval_merge(T)
                exclude_t = []
                if len(T) != 0 and len(n_T) != 0:
                    exclude_t = interval_merge([T, n_T])
                if len(exclude_t) != 0:
                    T = Interval.diff(T, exclude_t)

                if len(T) != 0:
                    if not isinstance(rule.head, Atom):
                        tmp_D = defaultdict(lambda: defaultdict(list))
                        tmp_D[head_predicate][replaced_head_entity] = T
                        tmp_head = copy.deepcopy(rule.head)
                        tmp_head.set_entity(replaced_head_entity)
                        T = reverse_apply(tmp_head, tmp_D)

                    delta_new[head_predicate][replaced_head_entity] = delta_new[head_predicate][
                                                                          replaced_head_entity] + T

        elif global_literal_index == visited:
            ground_body(global_literal_index+1, visited, delta, context)

        else:
            current_literal = copy.deepcopy(literals[global_literal_index])
            if not isinstance(current_literal, BinaryLiteral):
                if current_literal.get_predicate() in ["Bottom", "Top"]:
                    ground_body(global_literal_index+1, delta, context)
                else:
                    for tmp_entity, tmp_context in ground_generator(current_literal, context, D, D_index):
                        tmp_delata = {global_literal_index: [tmp_entity]}
                        # check whether it is needed to continue to do the evaluation
                        if current_literal.get_predicate() in non_predicates:
                            if not contain_variable_after_replace(head_entity, {**context, **tmp_context}):  # head can be grounded
                                if inclusion_check(D, rule, global_literal_index, records=records, context={**context, **tmp_context}):
                                    continue

                        ground_body(global_literal_index+1,  visited, {**delta, **tmp_delata}, {**context, **tmp_context})

            else:
                left_predicate = current_literal.left_literal.get_predicate()
                right_predicate = current_literal.right_literal.get_predicate()

                if left_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_context in ground_generator(current_literal.right_literal, context,  D, D_index):
                        tmp_delta = {global_literal_index: [tmp_entity]}

                        if right_predicate.get_predicate() in non_predicates:
                            if not contain_variable_after_replace(head_entity,
                                                                  {**context, **tmp_context}):  # head can be grounded
                                if inclusion_check(D, rule, global_literal_index, records=records,
                                                   context={**context, **tmp_context}):
                                    continue

                        ground_body(global_literal_index + 1, visited, {**delta, **tmp_delta}, {**context, **tmp_context})

                elif right_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_interval, tmp_context in ground_generator(current_literal.left_literal, context, D, D_index):
                        tmp_delta = {global_literal_index: [tmp_entity]}
                        if left_predicate.get_predicate() in non_predicates:
                            if not contain_variable_after_replace(head_entity,
                                                                  {**context, **tmp_context}):  # head can be grounded
                                if inclusion_check(D, rule, global_literal_index, records=records,
                                                   context={**context, **tmp_context}):
                                    continue

                        ground_body(global_literal_index + 1,visited, {**delta, **tmp_delta}, {**context, **tmp_context})

                else:
                    for left_entity, left_interval, tmp_context1 in ground_generator(current_literal.left_literal, context, D):
                        for right_entity, right_interval, tmp_context2 in ground_generator(current_literal.right_literal,{**context, **tmp_context1}, D):
                            tmp_delta = {global_literal_index: [left_entity, right_entity]}
                            if right_predicate.get_predicate() in non_predicates and left_predicate in non_predicates:
                                if not contain_variable_after_replace(head_entity, {**context, **tmp_context1, **tmp_context2}):  # head can be grounded
                                    if inclusion_check(D, rule, global_literal_index, records=records,
                                                       context={**context, **tmp_context1, **tmp_context2}):
                                        continue

                            ground_body(global_literal_index + 1, visited, {**delta, **tmp_delta}, {**context, **tmp_context1, **tmp_context2})

    for predicate in delta_old:
        for i in range(len(rule.body)):
            if not isinstance(literals[i], BinaryLiteral):
                if literals[i].get_predicate() == predicate:
                    break
            else:
                left_predicate = literals[i].left_atom.get_predicate()
                right_predicate = literals[i].right_atom.get_predicate()
                if left_predicate == predicate:
                    break
                elif right_predicate == predicate:
                    break
        else:
            continue
        for entity in delta_old[predicate]:
            for i in range(len(rule.body)):
                if isinstance(literals[i], BinaryLiteral):
                    left_predicate = literals[i].left_atom.get_predicate()
                    right_predicate = literals[i].right_atom.get_predicate()

                    if left_predicate == predicate:
                        context = dict()
                        for term1, term2 in zip(literals[i].left_atom.get_entity(), entity):
                            if term1.type == "constant" and term1.name != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in context and context[
                                term1.name] != term2.name:
                                break
                            else:
                                if term1.type == "variable":
                                    context[term1.name] = term2.name

                        if right_predicate in ["Bottom", "Top"]:
                            ground_body(0, i, {i: [entity]}, context)

                        else:
                            for tmp_entity, tmp_interval, tmp_context in ground_generator(literals[i].right_atom,
                                    context, D, D_index):
                                ground_body(0, i, {i: [entity, tmp_entity]}, {**context, **tmp_context})

                    elif right_predicate == predicate:
                        context = dict()
                        for term1, term2 in zip(literals[i].right_atom.get_entity(), entity):
                            if term1.type == "constant" and term1.name != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in context and context[
                                term1.name] != term2.name:
                                break
                            else:
                                if term1.type == "variable":
                                    context[term1.name] = term2.name
                        if left_predicate in ["Bottom", "Top"]:
                            ground_body(0, i, {i: [entity]}, context)
                        else:
                            for tmp_entity, tmp_interval, tmp_context in ground_generator(literals[i].left_atom, context, D, D_index):
                                ground_body(0, i, {i: [tmp_entity, entity]}, {**context, **tmp_context})

                else:
                    if literals[i].get_predicate() == predicate:
                        context = dict()
                        for term1, term2 in zip(literals[i].get_entity(), entity):
                            if term1.type == "constant" and term1.name != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                                break
                            else:
                                if term1.type == "variable":
                                    context[term1.name] = term2.name

                        # check whether it is needed to continue to do the evaluation
                        if literals[i].get_predicate() in non_predicates:
                            if not contain_variable_after_replace(head_entity, context): # head can be grounded
                                if inclusion_check(D, rule, i, records=records, context=context, single=True):
                                    continue

                        ground_body(0, i,  {i: [entity]}, context)


if __name__ == "__main__":
    from meteor_reasoner.utils.operate_dataset import print_dataset
    from meteor_reasoner.utils.loader import *
    from meteor_reasoner.materialization.materialize import seminaive_combine

    raw_program = [
        "P(X,Y):-Diamondminus[0,1]P(X,Y)",
        "R(Y):-Q(Y),H(Z),P(X,Y)",
        "Q(X):-K(X,Y)"
    ]
    raw_data = [
        "P(a,b)@0",
        "K(b,c)@[1,2]",
        "H(c)@[2,5]"
    ]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    program = load_program(raw_program)
    records = dict()
    non_predicates = ["H", "Q"]
    delta_old = D
    i = 0
    while i < 5:
        i += 1
        delta_new = defaultdict(lambda: defaultdict(list))
        for rule in program:
            optimized_seminaive_join(rule, D=D, delta_old=delta_old, delta_new=delta_new, D_index=D_index, non_predicates=non_predicates, records=records)
        print("{} iteration new:".format(i))
        print_dataset(delta_new)

        delta_old = defaultdict(lambda: defaultdict(list))
        fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)

    exit()
















import logging


def get_min_max_rational(D):
    min_val = float("inf")
    max_val = float("-inf")
    for predicate in D:
        for entity in D[predicate]:
            for intv in D[predicate][entity]:
                min_val = min(min_val, intv.left_value)
                max_val = max(max_val, intv.right_value)
    return min_val, max_val


def print_dataset(D):
    """
    Print all facts in D, the outputing form
    Args:
        D (a dictionary object):

    Returns:

    """
    for predicate in D:
        for entity, intervals in D[predicate].items():
            for interval in intervals:
                if len(entity) == 1 and entity[0].name == "nan":
                     print(predicate+"@"+str(interval))
                else:
                    print(predicate + "(" + ",".join([item.name for item in entity]) + ")@" + str(interval))


def save_dataset_to_file(filename, D):
    """
       Print all facts in D, the outputing form
       Args:
           D (a dictionary object):

       Returns:

       """
    filewriter = open(filename, "w")
    for predicate in D:
        for entity, intervals in D[predicate].items():
            for interval in intervals:
                if len(entity) == 1 and entity[0].name == "nan":
                    filewriter(predicate + "@" + str(interval)+"\n")
                else:
                    filewriter.write(predicate + "(" + ",".join([item.name for item in entity]) + ")@" + str(interval)+"\n")

def return_dataset(D):
    """
    Print all facts in D, the outputing form
    Args:
        D (a dictionary object):

    Returns:

    """
    res = ""
    for predicate in D:
        if "TMP" in predicate:
            continue
        if type(D[predicate]) == list:
            for interval in D[predicate]:
                res += predicate+"@"+str(interval) +"<br/>"
        else:
            for entity, intervals in D[predicate].items():
                entity_name = ",".join([item.name for item in entity if item.name != "nan"])
                for interval in intervals:
                    if entity_name != "":
                       res += predicate+"("+ entity_name + ")@"+str(interval) + "<br/>"
                    else:
                        res += predicate + "@" + str(interval) + "<br/>"

    return res


def yield_dataset(D):
    """
    Print all facts in D, the outputing form
    Args:
        D (a dictionary object):

    Returns:

    """
    res = ""
    for predicate in D:
        if "TMP" in predicate:
            continue
        if type(D[predicate]) == list:
            for interval in D[predicate]:
                res += predicate+"@"+str(interval) +"<br/>"
        else:
            for entity, intervals in D[predicate].items():
                entity_name = ",".join([item.name for item in entity if item.name != "nan"])
                for interval in intervals:
                    if entity_name != "":
                       yield predicate+"("+ entity_name + ")@"+str(interval)
                    else:
                        yield predicate + "@" + str(interval)

    return res



def print_predicate(predicate, D, num=100000):
    """
    Print all fact with a specified predicate.
    Args:
        predicate (str):
        D (a dictionary object):

    Returns:

    """
    if predicate in D:
        cnt = 0
        if type(D[predicate]) == list:
            for interval in D[predicate]:
                cnt += 1
                logging.info(predicate + "@" + str(interval))
                if cnt > num:
                    break
        else:
            for entity, intervals in D[predicate].items():
                for interval in intervals:
                    cnt += 1
                    if cnt < num:
                        logging.info(predicate + "(" + ",".join([item.name for item in entity]) + ")@" + str(interval))

        logging.info("The total number of the predicate({}) is {}".format(predicate, cnt))

    else:
        print("{} does not exist in D!".format(predicate))


def save_predicate(predicate, D, outfilename="result.txt"):
    """
    Print all fact with a specified predicate.
    Args:
        predicate (str):
        D (a dictionary object):

    Returns:

    """
    with open(outfilename, "w") as file:
        if predicate in D:
            if type(D[predicate]) == list:
                for interval in D[predicate]:
                    file.write(predicate + "@" + str(interval))
                    file.write("\n")
            else:
                for entity, intervals in D[predicate].items():
                    for interval in intervals:
                        file.write(predicate + "(" + ",".join([item.name for item in entity]) + ")@" + str(interval))
                        file.write("\n")
        else:
            print("{} does not exist in D!".format(predicate))


def save_dataset(D, outfilename):
    with open(outfilename, "w") as file:
        for predicate in D:
            if type(D[predicate]) == list:
                for interval in D[predicate]:
                    file.write(predicate + "@" + str(interval))
                    file.write("\n")
            else:
                for entity, intervals in D[predicate].items():
                    for interval in intervals:
                        file.write(predicate + "(" + ",".join([item.name for item in entity]) + ")@" + str(interval))
                        file.write("\n")



from meteor_reasoner.classes.interval import *

def entail(fact, D):
    if fact.predicate not in D:
        return False
    else:
        if not fact.entity in D[fact.predicate]:
            return False
        else:
            intervals = D[fact.predicate][fact.entity]
            for interval in intervals:
                if Interval.inclusion(fact.interval, interval):
                    return True
            else:
                return False